import os

def read_file(filename: str) -> str:
    """Read the contents of a file."""
    if not os.path.exists(filename):
        raise FileNotFoundError(f"File {filename} does not exist.")
    return open(filename).read()

def read_files(files: list[str]) -> list[str]:
    """Read the contents of a list of files."""
    return {f: read_file(f) for f in files}

def list_files(path: str) -> list[str]:
    """List the files and directory in a directory."""
    return [f for f in os.listdir(path) if not f.startswith(".")]

def tree(path: str, depth: int=3) -> str:
    """Get a tree representation of the files and directories in a directory."""
    def _tree(path: str, depth: int, prefix: str = "") -> str:
        if depth < 0:
            return ""
        if not os.path.exists(path):
            return f"{prefix} [Error: Path does not exist]"
        if os.path.isfile(path):
            return f"{prefix} {os.path.basename(path)}"
        try:
            files = sorted([f for f in os.listdir(path) if not f.startswith(".")])
        except PermissionError:
            return f"{prefix} [Error: Permission denied]"
        result = f"{prefix} {os.path.basename(path)}/" if prefix else ""
        for i, f in enumerate(files):
            is_last = (i == len(files) - 1)
            new_prefix = f"{prefix} {'└──' if is_last else '├──'}"
            result += _tree(os.path.join(path, f), depth - 1, new_prefix)
        return result

    if os.path.isfile(path):
        return os.path.basename(path)
    return _tree(path, depth).rstrip()



import reflex as rx
import datetime
import json
import inspect
from typing import Any, Dict, Callable

from flexai.capability import AgentCapability
from flexai.message import Message, AIMessage, UserMessage


class ToolInvocation(rx.Base):
    """Structure to specify a tool and arguments to invoke it with."""

    tool: str
    params: Dict[str, Any]

class ToolResponse(rx.Base):
    """Structure to specify the result of a tool invocation."""

    invocation: ToolInvocation
    result: Any
    error: str | None


class Tool(rx.Base):
    description: str | None
    params: Dict[str, Any]
    return_type: Any
    _func: Callable

    @classmethod
    def parse_function_annotations(cls, func: Callable) -> Dict[str, str]:
        """Parse function annotations into a more readable format."""
        signature = inspect.signature(func)
        params = {
            name: param.annotation.__name__ if hasattr(param.annotation, '__name__') else 'No annotation'
            for name, param in signature.parameters.items()
        }
        return_type = signature.return_annotation.__name__ if hasattr(signature.return_annotation, '__name__') else 'No annotation'
        description = inspect.getdoc(func) or 'No description'
        return {
            "name": func.__name__,
            "description": description,
            "params": params,
            "return_type": return_type
        }

    @classmethod
    def from_function(cls, func: Callable):
        """Create a tool from a function."""
        annotations = cls.parse_function_annotations(func)
        return cls(
            **annotations,
            _func=func,
        )

def send_message(message: str) -> None:
    """Send a final message to the user. This should be done after all internal processing is completed."""
    print(f"Sending message: {message}")


class ToolUse(AgentCapability):
    """Give an agent the ability to use tools."""

    def __init__(self, tools: list[Callable]):
        self.tools = [Tool.from_function(tool) for tool in tools]

    def init_agent(self, agent):
        agent.output_model = ToolInvocation

    async def modify_prompt(self, prompt: str) -> str:
        tool_info = "\n".join([json.dumps([tool.dict(exclude={"_func"}) for tool in self.tools], indent=2)])
        return f"""The current date and time is: {datetime.datetime.now()}.
{prompt}
You have the following tools available:
{tool_info}
Make sure your response includes only one of these valid tools.
"""

    async def modify_response(self, agent, messages, response: AIMessage) -> AIMessage:
        assert isinstance(response, ToolInvocation), f"Expected a ToolInvocation response, got {response}"
        # Base case
        if response.tool == "send_message":
            yield AIMessage(content=response.params["message"])
            return
        ai_message = AIMessage(content=json.dumps(response.dict(), indent=2))
        messages.append(ai_message)
        print(ai_message)

        yield ai_message
        tool = [tool for tool in self.tools if tool.name == response.tool][0]
        # Invoke the tool.
        try:
            result = tool._func(**response.params)
        except Exception as e:
            result = str(e)
        user_message = UserMessage(content=f"Tool invocation response: {result}")
        messages.append(user_message)
        yield user_message
        async for msg in agent.stream(messages):
            yield msg


from __future__ import annotations

from typing import TYPE_CHECKING
from flexai import llm, message
from flexai.tool import ToolInvocation

if TYPE_CHECKING:
    from flexai.capability import AgentCapability


class Agent:
    def __init__(self, prompt: str = "", capabilities: list[AgentCapability] = None):
        self.llm = llm.llm
        self.prompt = prompt
        self.capabilities = capabilities or []

    async def modify_messages(self, messages: list[message.Message]):
        for capability in self.capabilities:
            messages = await capability.modify_messages(messages)
        return messages

    async def get_system_message(self) -> message.SystemMessage:
        system = self.prompt
        for capability in self.capabilities:
            system = await capability.modify_prompt(system)
        return message.SystemMessage(content=system)

    async def stream(self, messages: list[message.Message]):
        messages = await self.modify_messages(messages)
        system = await self.get_system_message()
        response = await self.llm.get_structured_response(messages, model=ToolInvocation, system=system)
        for capability in self.capabilities:
            async for message in capability.modify_response(self, messages, response):
                if message is not None:
                    yield message

    async def get_response(self, messages: list[message.Message]):
        system = await self.get_system_message()
        messages = await self.modify_messages(messages)
        return await self.llm.get_chat_response(messages, system=system)


async def name_conversation(messages: list[message.Message]) -> str:
    assert len(messages) > 0, "No messages provided."
    first_message = messages[0].content
    prompt = f"""Given the following first message to this conversation, give it a short, succinct title: {first_message}.\n Include only the title, nothing else. Do not include any text other than the new title."""
    return (await llm.llm.get_chat_response([message.UserMessage(content=prompt)])).content


from flexai.message import Message, AIMessage


class AgentCapability:
    """Base class to define cognitive capabilities of an agent."""

    async def modify_prompt(self, prompt: str) -> str:
        """Modify the system prompt."""
        return prompt

    async def modify_messages(self, messages: list[Message]) -> list[Message]:
        """Modify the messages before sending them to the LLM."""
        return messages

    async def modify_response(self, agent, messages, response: AIMessage) -> AIMessage:
        yield None



from flexai.capability import AgentCapability
from flexai.message import Message, SystemMessage
from flexai.llm import llm


class TruncateMessages(AgentCapability):
    """Truncate the input messages to the LLM."""

    def __init__(self, max_messages: int):
        self.max_messages = max_messages

    async def modify_messages(self, messages: list[Message]) -> list[Message]:
        return messages[-self.max_messages:]


class UseEmotion(AgentCapability):
    """Add emotional context to the agent's responses."""

    def __init__(self):
        self.current_emotion = "I am excited to start helping!"

    async def modify_prompt(self, prompt: str) -> str:
        return f"""{prompt}

Your current emotional state is: {self.current_emotion}
Respond according to the emotional state.
"""


    async def modify_messages(self, messages: list[Message]) -> list[Message]:
        system = SystemMessage(content=f"""You are monitoring the emotional state of the ai agent.

        The current emotional state is: {self.current_emotion}

        Based on the current conversation, update the emotional state of the ai agent.
        Be descriptive but concise. Include only the new emotional state in your response, nothing else.
        """)
        self.emotional_state = (await llm.get_chat_response(system=system, messages=messages)).content
        print(f"New emotional state: {self.emotional_state}")
        return messages



import datetime
import reflex as rx
from sqlmodel import Column, DateTime, Field, func, select


class Conversation(rx.Model, table=True):
    """A conversation between the user and the AI."""

    # The name of the conversation.
    name: str

    # The timestamp when the conversation was created.
    timestamp: datetime.datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )


class Message(rx.Model, table=True):
    """A message to send to the agent."""

    # The role of the message (user, system, AI, tool).
    role: str

    # The content of the message.
    content: str

    # The type of data in the content.
    type: str = "text"

    # The timestamp when the message was created.
    timestamp: datetime.datetime = Field(
        sa_column=Column(DateTime(timezone=True), server_default=func.now()),
    )

    # The conversation ID the message belongs to.
    conversation_id: int | None = Field(default=None, foreign_key="conversation.id")

    def to_llm_message(self) -> dict:
        """Convert the message to a format that can be sent to the LLM."""
        return {
            "role": self.role,
            "content": self.content,
        }


def UserMessage(**kwargs):
    return Message(role="user", **kwargs)


def SystemMessage(**kwargs):
    return Message(role="system", **kwargs)


def AIMessage(**kwargs):
    return Message(role="assistant", **kwargs)


# class UserMessage(Message):
#     """A message from the user."""

#     role = "user"


# class SystemMessage(Message):
#     """A message from the system."""

#     role = "system"


# class AIMessage(Message):
#     """A message from the AI."""

#     role = "assistant"


# class ToolMessage(Message):
#     """A message from a tool."""

#     role = "tool"


def get_conversations(offset: int = 0, limit: int = 10):
    """Get a list of converations sorted from newest to oldest."""
    with rx.session() as session:
        return session.exec(
            select(Conversation)
            .order_by(Conversation.id.desc())
            .offset(offset)
            .limit(limit)
        ).all()


def get_messages(conversation_id: int, offset: int = 0, limit: int = 100):
    """Get the messages in a conversation."""
    with rx.session() as session:
        return session.exec(
            select(Message)
            .filter(Message.conversation_id == conversation_id)
            .order_by(Message.id)
            .offset(offset)
            .limit(limit)
        ).all()



from __future__ import annotations

import os
from typing import AsyncGenerator, Type

import reflex as rx
from flexai import message


class Client:
    """A base class for language models."""

    async def stream_chat_lines(
        self,
        messages: list[message.Message],
        system: message.SystemMessage | None = None,
    ) -> AsyncGenerator[str, None]:
        """Stream the response from the chat model line by line.

        Args:
            messages: The messages to send to the model.

        Returns:
            An async generator yielding the response from the model line by line.
        """
        buffer = ""
        async for delta in self.stream_chat_response(messages, system=system):
            buffer += delta
            while "\n" in buffer:
                line, buffer = buffer.split("\n", 1)
                yield line + "\n"


def to_llm_message(messages: list[message.Message]) -> list[dict]:
    """Convert the messages to a format that can be sent to the LLM.

    Args:
        messages: The messages to convert.

    Returns:
        The messages in a format that can be sent to the LLM.
    """
    return [m.to_llm_message() for m in messages]


class AnthropicClient(Client):
    def __init__(
        self, model=os.getenv("ANTHROPIC_MODEL", "claude-3-5-sonnet-20240620")
    ):
        from anthropic import AsyncAnthropic

        self.client = AsyncAnthropic()
        self.model = model
        self.max_tokens = 4096

    async def get_chat_response(
        self,
        messages: list[message.Message],
        system: message.SystemMessage | None = None,
    ) -> message.AIMessage:
        """Get the response from the chat model.

        Args:
            messages: The messages to send to the model.

        Returns:
            The response from the model.
        """
        system = system or message.SystemMessage(content="")
        response = (
            (
                await self.client.messages.create(
                    max_tokens=self.max_tokens,
                    messages=to_llm_message(messages),
                    model=self.model,
                    system=system.content or "",
                )
            )
        )
        print(response)
        content = response.content[0].text
        return message.AIMessage(content=content)


    async def stream_chat_response(
        self,
        messages: list[message.Message],
        system: message.SystemMessage | None = None,
    ) -> AsyncGenerator[message.AIMessage, None]:
        """Stream the response from the chat model.

        Args:
            messages: The messages to send to the model.

        Returns:
            An async generator yielding the response from the model.
        """
        system = system or message.SystemMessage(content="")
        ai_message = message.AIMessage(content="")
        async with self.client.messages.stream(
            max_tokens=self.max_tokens,
            messages=to_llm_message(messages),
            model=self.model,
            system=system.content or "",
        ) as stream:
            async for text in stream.text_stream:
                ai_message.content += text
                yield ai_message

    async def get_structured_response(
        self, messages: list[message.Message], model: Type[rx.Base], system: message.SystemMessage | None = None
    ) -> Type[rx.Base]:
        """Get the structured response from the chat model.

        Args:
            messages: The messages to send to the model.
            model: The model to use for the response.

        Returns:
            The structured response from the model.
        """
        system = message.SystemMessage(content=f"""{system or ""}

Return your answer according to the 'properties' of the following schema:
{model.schema()}

Return only the JSON object with the properties filled in.
Do not include anything in your response other than the JSON object.
Do not begin your response with ```json or end it with ```.
"""
        )
        print("final system")
        print(system.content)
        response = await self.get_chat_response(messages, system=system)
        print(response)
        try:
            obj = model.parse_raw(response.content)
        except Exception as e:
            # Try again, printing the exception.
            messages = messages + [
                response,
                message.UserMessage(content=f"There was an error while parsing. Make sure to only include the JSON. Error: {e}")
            ]
            return await self.get_structured_response(messages, model=model, system=system)
        return obj


llm = AnthropicClient()
